﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using RecrownedAthenaeum.Camera;
using RecrownedAthenaeum.ContentSystem;
using RecrownedAthenaeum.Data;
using RecrownedAthenaeum.Input;
using RecrownedAthenaeum.Persistence;
using RecrownedAthenaeum.ScreenSystem;
using RecrownedAthenaeum.SpecialTypes;
using RecrownedAthenaeum.UI.SkinSystem;
using RecrownedAthenaeum.UI.SkinSystem.Definitions;
using RhythmBullet.Audio;
using RhythmBullet.ContentResolvers;
using RhythmBullet.Preferences;
using RhythmBullet.Screens.MainMenu;
using System;
using System.Diagnostics;
using System.IO;

namespace RhythmBullet
{
    /// <summary>
    /// This is the main type for your game.
    /// </summary>
    public class RhythmBulletGame : Game
    {
        public const int WORLD_WIDTH = 4;
        public const int WORLD_HEIGHT = 3;
        public static int pixels_per_WU;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        public readonly ContentManagerController assets;
        
        readonly ResolutionContentResolver resolutionContentResolver;
        public PreferencesManager preferencesManager;
        private ScreenManager screenManager;
        public Camera2D Camera { get; private set; }
        private bool resizing;
        private bool initialLoadComplete;
        private MainScreen mainScreen;
        private Texture2D currentCursorTexture;
        internal readonly MusicController musicController;
        private SkinManager skinManager = new SkinManager();
        private ISkin Skin { get { return skinManager.Skin; } }

        public RhythmBulletGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
            assets = new ContentManagerController(Content);
            musicController = new MusicController();
            resolutionContentResolver = new ResolutionContentResolver();
            FontContentResolver fcr = new FontContentResolver(resolutionContentResolver);
            assets.contentPathModifier.Add(typeof(Texture2D), resolutionContentResolver);
            assets.contentPathModifier.Add(typeof(SpriteFont), fcr);
            preferencesManager = new PreferencesManager(
                Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + "/RhythmBullet",
                new General(),
                new Controls());
            preferencesManager.Load();
            Resolution resolution = preferencesManager.GetPreferences<General>().Resolution;
            graphics.PreferredBackBufferWidth = resolution.Width;
            graphics.PreferredBackBufferHeight = resolution.Height;
            RecrownedAthenaeum.Configuration.GraphicsDeviceManager = graphics;
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content. Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            General general = preferencesManager.GetPreferences<General>();
            Resolution resolution = general.Resolution;
            resolutionContentResolver.Width = resolution.Width;
            resolutionContentResolver.Height = resolution.Height;
            musicController.musicList.path = general.MusicDirectory;
            musicController.musicList.StartSearch();
            if ((general.skinName ) != null)
            {
                skinManager.LoadSkin(skinManager.ReadSkinData(general.skinDirectory + general.skinName), general.skinDirectory + general.skinName);
            }
            Debug.WriteLine("Initial setup complete.");
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            RecrownedAthenaeum.Configuration.Camera2D = new Camera2D();
            Camera = RecrownedAthenaeum.Configuration.Camera2D;

            screenManager = new ScreenManager();
            screenManager.ShowFirstScreenEvent += ShowFirstScreen;
            QueueContent();
            screenManager.Screen = new LoadingScreen(this, Content.Load<Texture2D>("RhythmBullet"), 0.7f);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// game-specific content.
        /// </summary>
        protected override void UnloadContent()
        {
            assets.UnloadAll();
            screenManager.Dispose();
            currentCursorTexture.Dispose();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.   
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            if (assets.Done)
            {
                if (resizing)
                {
                    assets.Update();
                }
                else if (!initialLoadComplete && CheckReadyForInitiate())
                {
                    PostLoad();
                }
            }
            else
            {
                assets.Update();
            }
            screenManager.UpdateCurrentScreen(gameTime, CheckReadyForInitiate());

            InputUtilities.Update();
            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            screenManager.DrawCurrentScreen(spriteBatch);
            base.Draw(gameTime);
        }

        public void Resize(int width, int height)
        {
            pixels_per_WU = height / WORLD_HEIGHT;
            resolutionContentResolver.Width = width;
            resolutionContentResolver.Height = height;
            graphics.PreferredBackBufferWidth = width;
            graphics.PreferredBackBufferHeight = height;
            resizing = true;
            assets.UnloadAll();
            QueueContent();
            screenManager.Resize(new LoadingScreen(this, Content.Load<Texture2D>("loading_ring"), 0.4f, true));
            UpdateCursor();
        }

        private void PostResize()
        {
            graphics.ApplyChanges();
            Camera.Apply();
            screenManager.PostResize();
        }

        private bool CheckReadyForInitiate()
        {
            bool ready = true;
            if (!assets.Done ||
                !musicController.musicList.Searched)
            {
                ready = false;
            }

            if (preferencesManager.GetPreferences<General>().skinName != null && !skinManager.MergingSkins)
            {
                ready = false;
            }
            return ready;
        }

        private void ShowFirstScreen(Screen Screen)
        {
            Screen.NextScreen = mainScreen;
        }

        private void QueueContent()
        {
            assets.Queue<Texture2D>("cursor", false);
            assets.Queue<Texture2D>("title");
            assets.Queue<Texture2D>("default_cover", false);
            assets.Queue<Texture2D>("backgrounds/mainBG");
            assets.Queue<TextureAtlas>("UI");
            assets.Queue<SpriteFont>("gasalt_regular");
            assets.Queue<SpriteFont>("darktech_ldr");
            assets.Queue<SpriteFont>("iron_maiden");
        }

        private void PostLoad()
        {
            Debug.WriteLine("Initial setup and loading complete.");
            SetUpDefaultSkin();
            UpdateCursor();

            mainScreen = new MainScreen(assets, Skin);
            initialLoadComplete = true;
        }

        private void UpdateCursor()
        {
            Texture2D texture = Skin.CursorTexture;
            int cursorSize = (int)(0.07f * graphics.PreferredBackBufferHeight);
            Debug.WriteLine("Cursor size length: " + cursorSize);
            RenderTarget2D renderTarget = new RenderTarget2D(GraphicsDevice, cursorSize, cursorSize);
            GraphicsDevice.SetRenderTarget(renderTarget);
            GraphicsDevice.Clear(Color.Transparent);
            spriteBatch.Begin();
            spriteBatch.Draw(texture, new Rectangle(0, 0, cursorSize, cursorSize), Color.White);
            spriteBatch.End();
            GraphicsDevice.SetRenderTarget(null);

            currentCursorTexture?.Dispose();
            currentCursorTexture = renderTarget;
            Mouse.SetCursor(MouseCursor.FromTexture2D(currentCursorTexture, currentCursorTexture.Width / 2, currentCursorTexture.Height / 2));
        }

        private void SetUpDefaultSkin()
        {
            Skin skin = new Skin(assets.Get<TextureAtlas>("UI"), assets.Get<Texture2D>("cursor"));
            skin.AddColor("default", Color.White);

            TextButtonSkinDefinition textButtonSkinDefinition = new TextButtonSkinDefinition("rectangle-button-down", "rectangle-button");
            textButtonSkinDefinition.disabledRegion = "rectangle-button-disabled";
            
            skin.AddDefinition(textButtonSkinDefinition);

            skin.Laminate();
            skinManager.BaseSkin = skin;
        }
    }
}
